﻿/*
  wavp_audio.c - wavp library
  Copyright (c) 2015 Mitsuhiro Matsuura.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/
/*
  Modified 20 Mar 2017 by Nozomu Fujita
   ・8bit /16bit 両データ対応
   ・WAV データのサンプリング周波数で PWM 再生周波数が変化しない様修正
*/

#include "iodefine.h"
#include "wavp_config.h"
#include "wavp_audio.h"

static int ch = 0;
static int databit = 0;
static int pwm_bit = 0;
static unsigned long pwm_clk = 0;
static unsigned long pwm_count;
static unsigned long sf = 0;
static unsigned long pwm_sf;

void wavp_pwmsf(unsigned long sf)
{
	pwm_sf = sf;
	return;
}

void wavp_pwmclk(unsigned long i)
{
	pwm_clk = i;
	return;
}

/* パラメータセット */
void wavp_audio_set(unsigned long p_sf, int p_databit, int p_ch)
{
	ch = p_ch;
	databit = p_databit;
	sf = p_sf;

	pwm_count = (unsigned short)(pwm_clk / pwm_sf);

	TDR00.tdr00 = pwm_count - 1;
	TDR01.tdr01 = pwm_count / 2 - 1;
	TDR02.tdr02 = pwm_count / 2 - 1;

	int n;
	for(n = 15; 0 <= n; n--)
	{
		if(pwm_count & (1 << n))
		{
			break;
		}
	}

	pwm_bit = n;
	return;
}

static unsigned short ringbuf[2][1024];		// ステレオ用2チャネル
volatile unsigned int decode_putp = 0, decode_getp = 0;

#define RINGBUF_SIZE_MAX (sizeof(ringbuf)/sizeof(unsigned short)/2)
#define p_inc(p, max) (((p + 1) >= max) ? 0 : (p + 1))

#define GET_NG 0
#define GET_OK 1

#define PUT_NG 0
#define PUT_OK 1

typedef short t_wav_data16;
#define WAV2DUTY16(w) ((w + (1L << databit - 1) >> databit - pwm_bit) + (pwm_count - (1L << pwm_bit)) / 2)
typedef unsigned char t_wav_data8;
#define WAV2DUTY8(w) ((w << pwm_bit - databit) + (pwm_count - (1L << pwm_bit)) / 2)

/* リングバッファに波形データを入れる */
static
int decode_put16(t_wav_data16 left, t_wav_data16 right)
{
	unsigned int putp = p_inc(decode_putp, RINGBUF_SIZE_MAX);
	int r = PUT_NG;

	// 空きがあるか
	if(putp != decode_getp)
	{
		// PWM DUTY値に変換
		ringbuf[0][decode_putp] = WAV2DUTY16(left);
		ringbuf[1][decode_putp] = WAV2DUTY16(right);

		decode_putp = putp;
		r = PUT_OK;
	}
	return r;
}

static
int decode_put8(t_wav_data8 left, t_wav_data8 right)
{
	unsigned int putp = p_inc(decode_putp, RINGBUF_SIZE_MAX);
	int r = PUT_NG;

	// 空きがあるか
	if(putp != decode_getp)
	{
		// PWM DUTY値に変換
		ringbuf[0][decode_putp] = WAV2DUTY8(left);
		ringbuf[1][decode_putp] = WAV2DUTY8(right);

		decode_putp = putp;
		r = PUT_OK;
	}
	return r;
}

void wavp_put(void *read_buffer, int size)
{
	int n;
	if (databit == 16) {
		t_wav_data16 audio_l, audio_r, *data = read_buffer;
	
		for(n = 0; n < size / ch / sizeof(t_wav_data16); n++)
		{
			audio_l = *data++;
			audio_r = (2 == ch) ? *data++ : audio_l;
	
			// リングバッファにデータを書きこむ
			// 空きが無ければ待つ
			while (PUT_OK != decode_put16(audio_l, audio_r));
		}
	} else if (databit == 8) {
		t_wav_data8 audio_l, audio_r, *data = read_buffer;
	
		for(n = 0; n < size / ch / sizeof(t_wav_data8); n++)
		{
			audio_l = *data++;
			audio_r = (2 == ch) ? *data++ : audio_l;
	
			// リングバッファにデータを書きこむ
			// 空きが無ければ待つ
			while (PUT_OK != decode_put8(audio_l, audio_r));
		}
	}
	return;
}

/* リングバッファから波形データが全部出終わるまで待つ */
void wavp_end(void)
{
	while(decode_putp != decode_getp);
}

/* リングバッファから波形データを取得 */
static
int decode_get(unsigned short *audio_l, unsigned short *audio_r)
{
	int r = GET_NG;

	// データがあるか
	if(decode_putp != decode_getp)
	{
		*audio_l = ringbuf[0][decode_getp];
		*audio_r = ringbuf[1][decode_getp];

		decode_getp = p_inc(decode_getp, RINGBUF_SIZE_MAX);
		r = GET_OK;
	}
	return r;
}

static
void audio(void)
{
	unsigned short audio_l, audio_r;

	if(GET_OK == decode_get(&audio_l, &audio_r))
	{
		// 取り出した波形データをPWMデューティにセット
		TDR01.tdr01 = audio_l;
		TDR02.tdr02 = audio_r;

		CHECK_UNDERFLOW();
	}
	return;
}

// タイマ0割り込み処理
void INT_TM00(void) __attribute__ ((interrupt));

void INT_TM00(void)
{
	static unsigned long sf_count = 0;

	sf_count += sf;
	while (sf_count >= pwm_sf) {
		audio();
		sf_count -= pwm_sf;
	}

	TMIF00 = 0;
}
