﻿/*
  wavp_play.c - wavp library
  Copyright (c) 2015 - 2016 Mitsuhiro Matsuura.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/
/*
  Modified 20 Mar 2017 by Nozomu Fujita
   ・8bit /16bit 両データ対応
   ・WAVP_PLAY_MESSAGE の値で sprintf() を使用しない様修正
*/

#include <stdio.h>
#include <string.h>
#include <wavp_b2ca.h>

#include "wavp_config.h"
#include "wavp_audio.h"
#include "wavp_init.h"
#include "wavp_play.h"

// 一時停止
static volatile int pause = 0;
int wavp_read_pause(void)
{
	return pause;
}

void wavp_pause(int command)
{
	pause = command;
	return;
}

// 再生中断
static volatile int skip = 0;
void wavp_skip(void)
{
	skip = 1;
	return;
}

// 再生制御
static int control(void)
{
	// 一時停止用ループ
	while(pause && !skip);
	pause = 0;

	// 中断
	if(skip)
	{
		skip = 0;
		return 0;
	}
	return 1;
}

#if WAVP_PLAY_MESSAGE
static char message[256];
#else
static char message[] = "";
#define sprintf(...)
#endif

// WAVファイル再生処理
char *wavp_play(const char *name)
{
	int res;
	B2CA fobj;
	unsigned int size;
	long chunk_size;
	short read_buffer[2*1024];

	char *filename = (char *)name;

	res = f_open(&fobj, filename);
	sprintf(message, "wavp_play error file[%s] line:%d, f_open:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;

	// "RIFF"チェック
	res = f_read(&fobj, (void *)read_buffer, 8, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	if(0 != strncmp((char *)read_buffer, "RIFF", 4))goto play_return;

	// "WAVE"チェック
	res = f_read(&fobj, (void *)read_buffer, 4, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	if(0 != strncmp((char *)read_buffer, "WAVE", 4))goto play_return;

	// "fmt "チェック
	res = f_read(&fobj, (void *)read_buffer, 4, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	while(0 != strncmp((char *)read_buffer, "fmt ", 4))
	{
		res = f_read(&fobj, (void *)read_buffer, 4, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
		unsigned long n = *(unsigned long *)read_buffer;
		if(sizeof(read_buffer) < n)goto play_return;
		res = f_read(&fobj, (void *)read_buffer, n, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
		res = f_read(&fobj, (void *)read_buffer, 4, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
	}

	// fmtチャンクデータサイズ取得
	res = f_read(&fobj, (void *)read_buffer, 4, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	unsigned long n = *(unsigned long *)read_buffer;
	if(0 == n)goto play_return;

	res = f_read(&fobj, (void *)read_buffer, n, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;

	// チャンネル数チェック
	// 2(ステレオ) と 1(モノラル) に対応
	int ch = *((unsigned short *)read_buffer + 1);
	sprintf(message, "wavp_play error file[%s] line:%d, ch:%d\r\n", __FILE__, __LINE__, ch);
	if((2 != ch) && (1 != ch))goto play_return;

	// サンプリング周波数
	unsigned long sf = *((unsigned long *)read_buffer + 1);
	sprintf(message, "wavp_play error file[%s] line:%d, sf:%d\r\n", __FILE__, __LINE__, sf);

	// ビット数チェック
	// 16bit と 8bit に対応
	int databit = *((unsigned short *)read_buffer + 7);
	sprintf(message, "wavp_play error file[%s] line:%d, bit:%d\r\n", __FILE__, __LINE__, databit);
	if(16 != databit && 8 != databit)goto play_return;

	// "data"チェック
	res = f_read(&fobj, (void *)read_buffer, 4, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	while(0 != strncmp((char *)read_buffer, "data", 4))
	{
		res = f_read(&fobj, (void *)read_buffer, 4, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
		unsigned long n = *(unsigned long *)read_buffer;
		if(sizeof(read_buffer) < n)goto play_return;
		res = f_read(&fobj, (void *)read_buffer, n, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
		res = f_read(&fobj, (void *)read_buffer, 4, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK) goto play_return;
	}

	// 波形データサイズ取得
	res = f_read(&fobj, (void *)read_buffer, 4, &size);
	sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
	if(res != FR_OK) goto play_return;
	chunk_size = *(long *)read_buffer;
	if(0 == chunk_size)goto play_return;

	// 読み出したパラメータをセット
	wavp_audio_set(sf, databit, ch);

	// 再生ループ
	while(control() && chunk_size > 0)
	{
		// 波形データをローカルバッファに読み出す
		res = f_read(&fobj, (void *)read_buffer, (sizeof(read_buffer) < chunk_size) ? sizeof(read_buffer) : chunk_size, &size);
		sprintf(message, "wavp_play error file[%s] line:%d, f_read:%d\r\n", __FILE__, __LINE__, res);
		if(res != FR_OK || 0 == size) break;

		// ローカルバッファからリングバッファにコピー
		wavp_put(read_buffer, (int)size);
		chunk_size -= size;
	}

	// 正常終了時はメッセージをクリア
	sprintf(message, "");

play_return:
	f_close(&fobj);

	// 波形出力終了待ち
	wavp_end();
	return message;
}
